@@ -5,8 +5,8 @@ module Options = struct
   type t =
     { with_deps : bool (* whether to compute direct dependencies between modules *)
     ; with_pps : bool
-        (* whether to include the dependencies to ppx-rewriters (that are
-           used at compile time) *)
+    (* whether to include the dependencies to ppx-rewriters (that are
+       used at compile time) *)
     }
 
   (* whether to sanitize absolute paths of workspace items, and their UIDs, to
@@ -74,7 +74,7 @@ module Descr = struct
       { for_intf : Dune_rules.Module_name.t list
           (* direct module dependencies for the interface *)
       ; for_impl : Dune_rules.Module_name.t list
-          (* direct module dependencies for the implementation *)
+      (* direct module dependencies for the implementation *)
       }
 
     (* Conversion to the [Dyn.t] type *)
@@ -388,13 +388,15 @@ module Crawl = struct
       >>= Dir_contents.ocaml
       >>| Ml_sources.modules_and_obj_dir ~for_:(Exe { first_exe })
     in
-    let pp_map =
+    let* pp_map =
+      let+ version =
+        let+ ocaml = Super_context.context sctx |> Context.ocaml in
+        ocaml.version
+      in
       Staged.unstage
-      @@
-      let version = (Super_context.context sctx |> Context.ocaml).version in
-      Preprocessing.pped_modules_map
-        (Preprocess.Per_module.without_instrumentation exes.buildable.preprocess)
-        version
+      @@ Preprocessing.pped_modules_map
+           (Preprocess.Per_module.without_instrumentation exes.buildable.preprocess)
+           version
     in
     let deps_of module_ =
       let module_ = pp_map module_ in
@@ -447,13 +449,16 @@ module Crawl = struct
             >>= Dir_contents.ocaml
             >>| Ml_sources.modules_and_obj_dir ~for_:(Library name)
           in
-          let pp_map =
+          let* pp_map =
+            let+ version =
+              let+ ocaml = Super_context.context sctx |> Context.ocaml in
+              ocaml.version
+            in
             Staged.unstage
-            @@
-            let version = (Super_context.context sctx |> Context.ocaml).version in
-            Preprocessing.pped_modules_map
-              (Preprocess.Per_module.without_instrumentation (Lib_info.preprocess info))
-              version
+            @@ Preprocessing.pped_modules_map
+                 (Preprocess.Per_module.without_instrumentation
+                    (Lib_info.preprocess info))
+                 version
           in
           let deps_of module_ =
             immediate_deps_of_module
@@ -518,10 +523,11 @@ module Crawl = struct
     dirs
     : Descr.Workspace.t Memo.t
     =
-    let sctx = Context_name.Map.find_exn scontexts (Context.name context) in
+    let context_name = Context.name context in
+    let sctx = Context_name.Map.find_exn scontexts context_name in
     let open Memo.O in
     let* dune_files =
-      Dune_load.Dune_files.eval conf.dune_files ~context
+      Dune_load.Dune_files.eval conf.dune_files ~context:context_name
       >>| List.filter ~f:(dune_file_is_in_dirs dirs)
     in
     let* exes, exe_libs =
@@ -577,7 +583,7 @@ let find_dir common dir =
 ;;
 
 let term : unit Term.t =
-  let+ common = Common.term
+  let+ builder = Common.Builder.term
   and+ what =
     Arg.(
       value
@@ -592,7 +598,7 @@ let term : unit Term.t =
   and+ format = Describe_format.arg
   and+ lang = Lang.arg
   and+ options = Options.arg in
-  let config = Common.init common in
+  let common, config = Common.init builder in
   let dirs =
     let args = "workspace" :: what in
     let parse =
@@ -627,7 +633,7 @@ let term : unit Term.t =
   let* setup = setup in
   let super_context = Import.Main.find_scontext_exn setup ~name:context_name in
   let context = Super_context.context super_context in
-  let findlib_paths = Context.findlib_paths context in
+  let* findlib_paths = Context.findlib_paths context in
   (* prefix directories with the workspace root, so that the
      command also works correctly when it is run from a
      subdirectory *)
