@@ -217,15 +217,13 @@ module Public_name = struct
     if is_opam_compatible l
     then Ok l
     else
-      let open Pp.O in
-      let descr =
-        Pp.text
-          "Public names are composed of an opam package name and optional dot-separated \
-           string suffixes."
-        ++ Pp.newline
-        ++ Pkg.description_of_valid_string
-      in
-      Error (User_error.make [ descr ])
+      Error
+        (User_error.make
+           [ Pp.text
+               "Public names are composed of an opam package name and optional \
+                dot-separated string suffixes."
+           ; Pkg.description_of_valid_string
+           ])
   ;;
 
   let of_name_exn name =
@@ -466,10 +464,11 @@ module Component = struct
           }
       in
       let test_target =
+        let test_name = "test_" ^ Dune_lang.Atom.to_string common.name in
         test
           { context = { context with dir = Path.relative dir "test" }
           ; options = ()
-          ; common
+          ; common = { common with name = Dune_lang.Atom.of_string test_name }
           }
       in
       let bin_target =
@@ -496,10 +495,11 @@ module Component = struct
           }
       in
       let test_target =
+        let test_name = "test_" ^ Dune_lang.Atom.to_string common.name in
         test
           { context = { context with dir = Path.relative dir "test" }
           ; options = ()
-          ; common
+          ; common = { common with name = Dune_lang.Atom.of_string test_name }
           }
       in
       lib_target @ test_target
