@@ -5,26 +5,18 @@ module Id = struct
     type t = Sexp.t
 
     let equal = Poly.equal
-
     let compare = Poly.compare
-
     let to_dyn s = Sexp.to_dyn s
   end
 
   include T
 
   let make s = s
-
   let to_sexp t = t
-
   let sexp = Conv.sexp
-
   let gen f = Conv.field "id" (f sexp)
-
   let required_field = gen Conv.required
-
   let optional_field = gen Conv.optional
-
   let hash = Poly.hash
 
   module C = Comparable.Make (T)
@@ -35,50 +27,55 @@ end
 module Version = struct
   type t = int * int
 
-  let latest = (3, 6)
+  let latest = 3, 12
 
   let sexp : t Conv.value =
     let open Conv in
     pair int int
+  ;;
 end
 
-module Method_name = struct
-  type t = string
+module Method = struct
+  module Name = struct
+    type t = string
 
-  let sexp : t Conv.value = Conv.string
+    let sexp : t Conv.value = Conv.string
 
-  module Map = String.Map
-  module Table = String.Table
-end
+    module Map = String.Map
+    module Table = String.Table
+  end
 
-module Method_version = struct
-  type t = int
+  module Version = struct
+    type t = int
 
-  let sexp = Conv.int
+    let sexp = Conv.int
 
-  module Set = Int.Set
-  module Map = Int.Map
+    module Set = Int.Set
+    module Map = Int.Map
+  end
 end
 
 module Call = struct
   type t =
-    { method_ : Method_name.t
+    { method_ : Method.Name.t
     ; params : Sexp.t
     }
 
   let to_dyn { method_; params } =
     let open Dyn in
-    record [ ("method_", String method_); ("params", Sexp.to_dyn params) ]
+    record [ "method_", String method_; "params", Sexp.to_dyn params ]
+  ;;
 
   let create ?(params = Sexp.List []) ~method_ () = { method_; params }
 
   let fields =
     let open Conv in
     let to_ (method_, params) = { method_; params } in
-    let from { method_; params } = (method_, params) in
-    let method_ = field "method" (required Method_name.sexp) in
+    let from { method_; params } = method_, params in
+    let method_ = field "method" (required Method.Name.sexp) in
     let params = field "params" (required sexp) in
     iso (both method_ params) to_ from
+  ;;
 end
 
 module Request = struct
@@ -98,6 +95,7 @@ module Response = struct
       | Invalid_request -> variant "Invalid_request" []
       | Code_error -> variant "Code_error" []
       | Connection_dead -> variant "Connection_dead" []
+    ;;
 
     type t =
       { payload : Sexp.t option
@@ -106,9 +104,7 @@ module Response = struct
       }
 
     let payload t = t.payload
-
     let kind t = t.kind
-
     let message t = t.message
 
     exception E of t
@@ -126,36 +122,39 @@ module Response = struct
       | Version_error { payload; message; since = _; until = _ } ->
         (* cwong: Should we even still include this? *)
         { message; payload = make_payload payload; kind = Code_error }
+    ;;
 
     let sexp =
       let open Conv in
       let id = field "payload" (optional sexp) in
       let message = field "message" (required string) in
       let kind =
-        field "kind"
+        field
+          "kind"
           (required
-             (enum
-                [ ("Invalid_request", Invalid_request)
-                ; ("Code_error", Code_error)
-                ]))
+             (enum [ "Invalid_request", Invalid_request; "Code_error", Code_error ]))
       in
       record
-        (iso (three id message kind)
+        (iso
+           (three id message kind)
            (fun (payload, message, kind) -> { payload; message; kind })
-           (fun { payload; message; kind } -> (payload, message, kind)))
+           (fun { payload; message; kind } -> payload, message, kind))
+    ;;
 
     let to_dyn { payload; message; kind } =
       let open Dyn in
       record
-        [ ("payload", option Sexp.to_dyn payload)
-        ; ("message", string message)
-        ; ("kind", dyn_of_kind kind)
+        [ "payload", option Sexp.to_dyn payload
+        ; "message", string message
+        ; "kind", dyn_of_kind kind
         ]
+    ;;
 
     let () =
       Printexc.register_printer (function
         | E e -> Some (Dyn.to_string (Dyn.variant "Response.E" [ to_dyn e ]))
         | _ -> None)
+    ;;
   end
 
   type t = (Sexp.t, Error.t) result
@@ -169,12 +168,14 @@ module Response = struct
       (function
         | Ok s -> case s ok
         | Error e -> case e error)
+  ;;
 
   let fields =
     let open Conv in
     let id = Id.required_field in
     let payload = field "result" (required result) in
     both id payload
+  ;;
 
   let to_dyn = Result.to_dyn Sexp.to_dyn Error.to_dyn
 end
@@ -183,7 +184,6 @@ module Protocol = struct
   type t = int
 
   let latest_version = 0
-
   let sexp = Conv.int
 end
 
@@ -196,99 +196,97 @@ module Initialize = struct
       }
 
     let dune_version t = t.dune_version
-
     let protocol_version t = t.protocol_version
-
     let id t = t.id
 
     let create ~id =
       let dune_version = Version.latest in
       let protocol_version = Protocol.latest_version in
       { dune_version; protocol_version; id }
+    ;;
 
     let method_name = "initialize"
 
     let sexp =
       let open Conv in
       let dune_version = field "dune_version" (required Version.sexp) in
-      let protocol_version =
-        field "protocol_version" (required Protocol.sexp)
-      in
+      let protocol_version = field "protocol_version" (required Protocol.sexp) in
       let id = Id.required_field in
       let to_ (dune_version, protocol_version, id) =
         { dune_version; protocol_version; id }
       in
       let from { dune_version; protocol_version; id } =
-        (dune_version, protocol_version, id)
+        dune_version, protocol_version, id
       in
       record (iso (three dune_version protocol_version id) to_ from)
+    ;;
 
     let of_call { Call.method_; params } ~version =
-      if String.equal method_ method_name then
-        Conv.of_sexp sexp ~version params
-        |> Result.map_error ~f:Response.Error.of_conv
-      else
+      if String.equal method_ method_name
+      then Conv.of_sexp sexp ~version params |> Result.map_error ~f:Response.Error.of_conv
+      else (
         let message = "initialize request expected" in
-        Error (Response.Error.create ~message ~kind:Invalid_request ())
+        Error (Response.Error.create ~message ~kind:Invalid_request ()))
+    ;;
 
     let to_call t =
       let params = Conv.to_sexp sexp t in
       { Call.method_ = "initialize"; params }
+    ;;
   end
 
   module Response = struct
     type t = unit
 
     let sexp = Conv.unit
-
     let create () = ()
-
     let to_response t = Conv.to_sexp sexp t
   end
 end
 
 module Version_negotiation = struct
   module Request = struct
-    type t = Menu of (string * int list) list
+    type t = Menu of (Method.Name.t * Method.Version.t list) list
 
     let method_name = "version_menu"
-
     let create menu = Menu menu
 
     let sexp =
       Conv.(
         iso
-          (list (pair string (list int)))
+          (list (pair Method.Name.sexp (list Method.Version.sexp)))
           (fun x -> Menu x)
           (function
             | Menu x -> x))
+    ;;
 
     let to_call t =
       let params = Conv.to_sexp sexp t in
       { Call.method_ = method_name; params }
+    ;;
 
     let of_call { Call.method_; params } ~version =
-      if String.equal method_ method_name then
-        Conv.of_sexp sexp ~version params
-        |> Result.map_error ~f:Response.Error.of_conv
-      else
+      if String.equal method_ method_name
+      then Conv.of_sexp sexp ~version params |> Result.map_error ~f:Response.Error.of_conv
+      else (
         let message = "version negotiation request expected" in
-        Error (Response.Error.create ~message ~kind:Invalid_request ())
+        Error (Response.Error.create ~message ~kind:Invalid_request ()))
+    ;;
   end
 
   module Response = struct
-    type t = Selected of (string * int) list
+    type t = Selected of (Method.Name.t * Method.Version.t) list
 
     let sexp =
       Conv.(
         iso
-          (list (pair string int))
+          (list (pair Method.Name.sexp Method.Version.sexp))
           (fun x -> Selected x)
           (function
             | Selected x -> x))
+    ;;
 
     let create x = Selected x
-
     let to_response t = Conv.to_sexp sexp t
   end
 end
@@ -310,6 +308,7 @@ module Persistent = struct
         (function
           | Packet p -> case p packet
           | Close_connection -> case () close_connection)
+    ;;
   end
 
   module In = struct
@@ -324,12 +323,11 @@ module Persistent = struct
       | New_connection -> variant "New_connection" []
       | Close_connection -> variant "Close_connection" []
       | Packet sexp -> variant "Packet" [ Sexp.to_dyn sexp ]
+    ;;
 
     let sexp =
       let open Conv in
-      let new_connection =
-        constr "new_connection" unit (fun () -> New_connection)
-      in
+      let new_connection = constr "new_connection" unit (fun () -> New_connection) in
       let packet = constr "packet" sexp (fun p -> Packet p) in
       let close_connection =
         constr "close_connection" unit (fun () -> Close_connection)
@@ -340,56 +338,59 @@ module Persistent = struct
           | New_connection -> case () new_connection
           | Packet p -> case p packet
           | Close_connection -> case () close_connection)
+    ;;
   end
 end
 
 module Packet = struct
-  module Query = struct
-    type t =
-      | Request of Request.t
-      | Notification of Call.t
+  type t =
+    | Request of Request.t
+    | Response of (Id.t * Response.t)
+    | Notification of Call.t
 
-    let sexp =
-      let open Conv in
-      let to_ (id, payload) =
-        match id with
-        | None -> Notification payload
-        | Some id -> Request (id, payload)
-      in
-      let from = function
-        | Request (id, payload) -> (Some id, payload)
-        | Notification payload -> (None, payload)
-      in
-      record
-        (iso
-           (let id = Id.optional_field in
-            both id Call.fields)
-           to_ from)
-  end
-
-  module Reply = struct
-    type t =
-      | Response of (Id.t * Response.t)
-      | Notification of Call.t
-
-    let sexp =
-      let f = function
-        | Either.Left (id, resp) -> Response (id, resp)
-        | Right y -> Notification y
-      in
-      let t = function
-        | Response (id, resp) -> Either.Left (id, resp)
-        | Notification r -> Right r
-      in
-      let open Conv in
-      record (iso (either Response.fields Call.fields) f t)
-  end
+  let sexp =
+    let open Conv in
+    let to_ (id, result, call) =
+      match id, result, call with
+      | Some id, None, Some call -> Request (id, call)
+      | None, None, Some call -> Notification call
+      | Some id, Some result, None -> Response (id, result)
+      | _, _, _ -> Conv.error (Parse_error { message = "invalid packet"; payload = [] })
+    in
+    let from = function
+      | Request (id, payload) -> Some id, None, Some payload
+      | Response (id, response) -> Some id, Some response, None
+      | Notification call -> None, None, Some call
+    in
+    record
+    @@ iso
+         (let id = Id.optional_field in
+          let result = field "result" (optional Response.result) in
+          let method_ = field "method" (optional Method.Name.sexp) in
+          let params = field "params" (optional sexp) in
+          let call =
+            iso
+              (both method_ params)
+              (fun (method_, params) ->
+                match method_, params with
+                | Some method_, Some params -> Some { Call.method_; params }
+                | None, None -> None
+                | Some _, None | None, Some _ ->
+                  Conv.error (Parse_error { message = "invalid call"; payload = [] }))
+              (function
+                | Some { Call.method_; params } -> Some method_, Some params
+                | None -> None, None)
+          in
+          three id result call)
+         to_
+         from
+  ;;
 end
 
 module Decl = struct
   type 'gen t =
-    { method_ : Method_name.t
-    ; key : 'gen Int.Map.t Univ_map.Key.t
+    { method_ : Method.Name.t
+    ; key : 'gen Method.Version.Map.t Univ_map.Key.t
     }
 
   module Generation = struct
@@ -409,24 +410,32 @@ module Decl = struct
   end
 
   module Request = struct
-    type ('req, 'resp) gen = Method_version.t * ('req, 'resp) Generation.t
+    type ('req, 'resp) gen = Method.Version.t * ('req, 'resp) Generation.t
 
-    let make_gen ~req ~resp ~upgrade_req ~downgrade_req ~upgrade_resp
-        ~downgrade_resp ~version =
+    let make_gen
+      ~req
+      ~resp
+      ~upgrade_req
+      ~downgrade_req
+      ~upgrade_resp
+      ~downgrade_resp
+      ~version
+      =
       ( version
       , Generation.T
-          { req
-          ; resp
-          ; upgrade_req
-          ; downgrade_req
-          ; upgrade_resp
-          ; downgrade_resp
-          } )
+          { req; resp; upgrade_req; downgrade_req; upgrade_resp; downgrade_resp } )
+    ;;
 
     let make_current_gen ~req ~resp ~version =
-      let id x = x in
-      make_gen ~req ~resp ~upgrade_req:id ~downgrade_req:id ~upgrade_resp:id
-        ~downgrade_resp:id ~version
+      make_gen
+        ~req
+        ~resp
+        ~upgrade_req:Fun.id
+        ~downgrade_req:Fun.id
+        ~upgrade_resp:Fun.id
+        ~downgrade_resp:Fun.id
+        ~version
+    ;;
 
     let gen_to_dyn _ = Dyn.String "<generation>"
 
@@ -440,33 +449,54 @@ module Decl = struct
     let make ~method_ ~generations =
       { generations
       ; decl =
-          { method_
-          ; key = Univ_map.Key.create ~name:method_ (Int.Map.to_dyn gen_to_dyn)
-          }
+          { method_; key = Univ_map.Key.create ~name:method_ (Int.Map.to_dyn gen_to_dyn) }
       }
+    ;;
 
+    let print_generation_list ~include_response generations =
+      List.iter generations ~f:(fun (version, Generation.T conv) ->
+        let conv_to_digest conv =
+          let sexp_string = Sexp.to_string (Conv.sexp_for_digest conv) in
+          if String.length sexp_string < 32
+          then sexp_string
+          else Digest.to_hex (Digest.string sexp_string)
+        in
+        let req = conv_to_digest conv.req in
+        let resp = conv_to_digest conv.resp in
+        if include_response
+        then Printf.printf "Version %d:\n  Request: %s\n  Response: %s\n" version req resp
+        else Printf.printf "Version %d: %s\n" version req)
+    ;;
+
+    let print_generations t = print_generation_list ~include_response:true t.generations
     let witness t = t.decl
   end
 
   module Notification = struct
-    type 'payload gen = Method_version.t * ('payload, unit) Generation.t
+    type 'payload gen = Method.Version.t * ('payload, unit) Generation.t
 
-    let make_gen (type a b) ~(conv : a Conv.value) ~(upgrade : a -> b)
-        ~(downgrade : b -> a) ~version : b gen =
-      let id x = x in
+    let make_gen
+      (type a b)
+      ~(conv : a Conv.value)
+      ~(upgrade : a -> b)
+      ~(downgrade : b -> a)
+      ~version
+      : b gen
+      =
       ( version
       , Generation.T
           { req = conv
           ; resp = Conv.unit
           ; upgrade_req = upgrade
           ; downgrade_req = downgrade
-          ; upgrade_resp = id
-          ; downgrade_resp = id
+          ; upgrade_resp = Fun.id
+          ; downgrade_resp = Fun.id
           } )
+    ;;
 
     let make_current_gen (type a) ~(conv : a Conv.value) ~version : a gen =
-      let id x = x in
-      make_gen ~conv ~upgrade:id ~downgrade:id ~version
+      make_gen ~conv ~upgrade:Fun.id ~downgrade:Fun.id ~version
+    ;;
 
     let gen_to_dyn _ = Dyn.String "<generation>"
 
@@ -480,15 +510,17 @@ module Decl = struct
     let make ~method_ ~generations =
       { generations
       ; decl =
-          { method_
-          ; key = Univ_map.Key.create ~name:method_ (Int.Map.to_dyn gen_to_dyn)
-          }
+          { method_; key = Univ_map.Key.create ~name:method_ (Int.Map.to_dyn gen_to_dyn) }
       }
+    ;;
+
+    let print_generations t =
+      Request.print_generation_list ~include_response:false t.generations
+    ;;
 
     let witness t = t.decl
   end
 
   type ('a, 'b) request = ('a, 'b) Request.t
-
   type 'a notification = 'a Notification.t
 end
