@@ -235,7 +235,7 @@ end = struct
 
   (* The current version of the rule digest scheme. We should increment it when
      making any changes to the scheme, to avoid collisions. *)
-  let rule_digest_version = 16
+  let rule_digest_version = 20
 
   let compute_rule_digest
     (rule : Rule.t)
@@ -264,14 +264,13 @@ end = struct
       , sandbox_mode
       , Dep.Facts.digest deps ~env
       , file_targets @ dir_targets
-      , Option.map rule.context ~f:(fun c -> Context_name.to_string c.name)
       , Action.for_shell action
       , can_go_in_shared_cache
       , List.map locks ~f:Path.to_string
       , Execution_parameters.action_stdout_on_success execution_parameters
       , Execution_parameters.action_stderr_on_success execution_parameters
       , Execution_parameters.workspace_root_to_build_path_prefix_map execution_parameters
-      )
+      , Execution_parameters.expand_aliases_in_sandbox execution_parameters )
     in
     Digest.generic trace
   ;;
@@ -314,7 +313,6 @@ end = struct
     ~action
     ~deps
     ~loc
-    ~(context : Build_context.t option)
     ~execution_parameters
     ~sandbox_mode
     ~dir
@@ -369,6 +367,7 @@ end = struct
         Action.chdirs action
         |> Path.Build.Set.iter ~f:(fun p -> Path.mkdir_p (Path.build p)))
     in
+    let context = Build_context.of_build_path dir in
     let root =
       match context with
       | None -> Path.Build.root
@@ -380,53 +379,46 @@ end = struct
          | None -> root
          | Some sandbox -> Sandbox.map_path sandbox root)
     in
-    let* exec_result =
-      Fiber.collect_errors
-      @@ fun () ->
-      with_locks locks ~f:(fun () ->
-        let build_deps deps = Memo.run (build_deps deps) in
-        let* action_exec_result =
-          let input =
-            { Action_exec.root
-            ; context
-            ; env
-            ; targets = Some targets
-            ; rule_loc = loc
-            ; execution_parameters
-            ; action
-            }
-          in
-          match (Build_config.get ()).action_runner input with
-          | None -> Action_exec.exec input ~build_deps
-          | Some runner -> Action_runner.exec_action runner input
-        in
-        let* action_exec_result = Action_exec.Exec_result.ok_exn action_exec_result in
-        let+ produced_targets =
-          match sandbox with
-          | None -> Targets.Produced.produced_after_rule_executed_exn ~loc targets
-          | Some sandbox ->
-            (* The stamp file for anonymous actions is always created outside
-               the sandbox, so we can't move it. *)
-            let should_be_skipped =
-              match rule_kind with
-              | Normal_rule -> fun (_ : Path.Build.t) -> false
-              | Anonymous_action { stamp_file; _ } -> Path.Build.equal stamp_file
+    Fiber.finalize
+      ~finally:(fun () ->
+        match sandbox with
+        | Some sandbox -> Sandbox.destroy sandbox
+        | None ->
+          Pending_targets.remove targets;
+          Fiber.return ())
+      (fun () ->
+        with_locks locks ~f:(fun () ->
+          let build_deps deps = Memo.run (build_deps deps) in
+          let* action_exec_result =
+            let input =
+              { Action_exec.root
+              ; context (* can be derived from the root *)
+              ; env
+              ; targets = Some targets
+              ; rule_loc = loc
+              ; execution_parameters
+              ; action
+              }
             in
-            Sandbox.move_targets_to_build_dir sandbox ~loc ~should_be_skipped ~targets
-        in
-        { Exec_result.produced_targets; action_exec_result })
-    in
-    let* () =
-      match sandbox with
-      | Some sandbox -> Sandbox.destroy sandbox
-      | None ->
-        (* All went well, these targets are no longer pending. *)
-        Pending_targets.remove targets;
-        Fiber.return ()
-    in
-    match exec_result with
-    | Ok res -> Fiber.return res
-    | Error l -> Fiber.reraise_all l
+            match (Build_config.get ()).action_runner input with
+            | None -> Action_exec.exec input ~build_deps
+            | Some runner -> Action_runner.exec_action runner input
+          in
+          let* action_exec_result = Action_exec.Exec_result.ok_exn action_exec_result in
+          let+ produced_targets =
+            match sandbox with
+            | None -> Targets.Produced.produced_after_rule_executed_exn ~loc targets
+            | Some sandbox ->
+              (* The stamp file for anonymous actions is always created outside
+                 the sandbox, so we can't move it. *)
+              let should_be_skipped =
+                match rule_kind with
+                | Normal_rule -> fun (_ : Path.Build.t) -> false
+                | Anonymous_action { stamp_file; _ } -> Path.Build.equal stamp_file
+              in
+              Sandbox.move_targets_to_build_dir sandbox ~loc ~should_be_skipped ~targets
+          in
+          { Exec_result.produced_targets; action_exec_result }))
   ;;
 
   let promote_targets ~rule_mode ~dir ~targets ~promote_source =
@@ -438,7 +430,7 @@ end = struct
   ;;
 
   let execute_rule_impl ~rule_kind rule =
-    let { Rule.id = _; targets; dir; context; mode; action; info = _; loc } = rule in
+    let { Rule.id = _; targets; dir; mode; action; info = _; loc } = rule in
     (* We run [State.start_rule_exn ()] entirely for its side effect, so one
        might be tempted to use [Memo.of_non_reproducible_fiber] here but that is
        wrong, because that would force us to rerun [execute_rule_impl] on every
@@ -447,9 +439,10 @@ end = struct
     let head_target = Targets.Validated.head targets in
     let* execution_parameters =
       match Dpath.Target_dir.of_target dir with
-      | Regular (With_context (_, dir)) | Anonymous_action (With_context (_, dir)) ->
+      | Regular (With_context (_, _)) | Anonymous_action (With_context (_, _)) ->
         (Build_config.get ()).execution_parameters ~dir
-      | _ -> Execution_parameters.default
+      | Anonymous_action Root | Regular Root | Invalid _ ->
+        Code_error.raise "invalid dir for rule execution" [ "dir", Path.Build.to_dyn dir ]
     in
     (* Note: we do not run the below in parallel with the above: if we fail to
        compute action execution parameters, we have no use for the action and
@@ -477,22 +470,10 @@ end = struct
       let* () = Targets.maybe_async (fun () -> Path.mkdir_p (Path.build dir)) in
       let is_action_dynamic = Action.is_dynamic action.action in
       let sandbox_mode =
-        match Action.is_useful_to_sandbox action.action with
-        | Clearly_not ->
-          if Sandbox_config.mem action.sandbox Sandbox_mode.none
-          then Sandbox_mode.none
-          else
-            User_error.raise
-              ~loc
-              [ Pp.text
-                  "Rule dependencies are configured to require sandboxing, but the rule \
-                   has no actions that could potentially require sandboxing."
-              ]
-        | Maybe ->
-          select_sandbox_mode
-            ~loc
-            action.sandbox
-            ~sandboxing_preference:config.sandboxing_preference
+        select_sandbox_mode
+          ~loc
+          action.sandbox
+          ~sandboxing_preference:config.sandboxing_preference
       in
       (* CR-someday amokhov: More [always_rerun] and [can_go_in_shared_cache]
          to [Rule_cache] too. *)
@@ -584,7 +565,6 @@ end = struct
                   ~action
                   ~deps
                   ~loc
-                  ~context
                   ~execution_parameters
                   ~sandbox_mode
                   ~dir
@@ -623,7 +603,7 @@ end = struct
           ~rule_mode:mode
           ~dir
           ~targets:produced_targets
-          ~promote_source:(config.promote_source context)
+          ~promote_source:config.promote_source
       in
       let+ () = State.incr_rule_done_exn () in
       produced_targets)
@@ -668,9 +648,8 @@ end = struct
       Path.Build.relative dir basename
     in
     let rule =
-      let { Rule.Anonymous_action.context; action = _; loc; dir = _; alias = _ } = act in
+      let { Rule.Anonymous_action.action = _; loc; dir = _; alias = _ } = act in
       Rule.make
-        ~context
         ~info:(if Loc.is_none loc then Internal else From_dune_file loc)
         ~targets:(Targets.File.create target)
         ~mode:Standard
@@ -735,9 +714,9 @@ end = struct
     let observing_facts = () in
     ignore observing_facts;
     let digest =
-      let { Rule.Anonymous_action.context
-          ; action = { action; env; locks; can_go_in_shared_cache; sandbox }
-          ; loc
+      let { Rule.Anonymous_action.action =
+              { action; env; locks; can_go_in_shared_cache; sandbox }
+          ; loc = _
           ; dir
           ; alias
           }
@@ -760,15 +739,10 @@ end = struct
         |> Env.Map.to_list
       in
       Digest.generic
-        ( Option.map context ~f:(fun c ->
-            (* Only looking at the context name is fishy, but it is in line
-               with what we do for build rules. *)
-            Context_name.to_string c.name)
-        , env
+        ( env
         , Dep.Set.digest deps
         , Action.for_shell action
         , List.map locks ~f:Path.to_string
-        , loc
         , dir
         , alias
         , capture_stdout
