@@ -13,9 +13,15 @@ end
 
 include struct
   open Dune_lang
-  module Package_name = Package_name
   module Action = Action
   module String_with_vars = String_with_vars
   module Pform = Pform
-  module Blang = Dune_lang.Blang
+  module Blang = Blang
+  module Slang = Slang
+end
+
+include struct
+  open Dune_engine
+  module Process = Process
+  module Display = Display
 end
