@@ -5,19 +5,6 @@ open Import
 let ( let+! ) = Memo.O.( let+ )
 let ( let*! ) = Memo.O.( let* )
 
-let as_in_build_dir ~what ~loc p =
-  match Path.as_in_build_dir p with
-  | Some p -> p
-  | None ->
-    User_error.raise
-      ~loc
-      [ Pp.textf
-          "%s %s is outside the build directory. This is not allowed."
-          what
-          (Path.to_string_maybe_quoted p)
-      ]
-;;
-
 let validate_target_dir ~targets_dir ~loc targets path =
   if Path.Build.(parent_exn path <> targets_dir)
   then
@@ -368,12 +355,12 @@ end = struct
       then
         Memo.return
           ( (let+ p = Expander.expand_path env sw in
-             as_in_build_dir ~what ~loc:(loc sw) p)
+             Expander0.as_in_build_dir ~what ~loc:(loc sw) p)
           , acc )
       else
         let+! p = Expander.No_deps.expand_path env sw in
         let loc = loc sw in
-        let p = as_in_build_dir p ~what ~loc in
+        let p = Expander0.as_in_build_dir p ~what ~loc in
         Action_builder.return p, { acc with file_targets = f acc.file_targets p loc }
     ;;
 
@@ -404,10 +391,7 @@ end = struct
                Action_builder.return (Ok (Path.relative dir s))
              | In_path ->
                Action_builder.of_memo
-                 (Artifacts.Bin.binary
-                    ~loc:(Some loc)
-                    (Expander.artifacts env.expander)
-                    s))
+                 (Artifacts.binary ~loc:(Some loc) (Expander.artifacts env.expander) s))
         in
         let prog = Result.map prog ~f:(Expander.map_exe env.expander) in
         let args = Value.L.to_strings ~dir args in
@@ -432,9 +416,19 @@ let rec expand (t : Dune_lang.Action.t) : Action.t Action_expander.t =
     prog, more_args @ args
   in
   match t with
-  | Run (prog, args) ->
-    let+ prog, args = expand_run prog args in
-    O.Run (prog, Array.Immutable.of_list args)
+  | Run args ->
+    let string_args =
+      List.filter_map args ~f:(function
+        | Slang.Literal sw -> Some sw
+        | _ -> None)
+    in
+    if List.length string_args < List.length args
+    then User_error.raise [ Pp.text "All arguments to \"run\" action must be strings" ];
+    (match string_args with
+     | prog :: args ->
+       let+ prog, args = expand_run prog args in
+       O.Run (prog, Array.Immutable.of_list args)
+     | [] -> User_error.raise [ Pp.text "\"run\" action must have at least one argument" ])
   | With_accepted_exit_codes (pred, t) ->
     let+ t = expand t in
     O.With_accepted_exit_codes (pred, t)
@@ -444,7 +438,7 @@ let rec expand (t : Dune_lang.Action.t) : Action.t Action_expander.t =
   | Chdir (fn, t) ->
     E.At_rule_eval_stage.path fn ~f:(fun dir ->
       A.chdir
-        (as_in_build_dir dir ~loc:(String_with_vars.loc fn) ~what:"Directory")
+        (Expander0.as_in_build_dir dir ~loc:(String_with_vars.loc fn) ~what:"Directory")
         (let+ t = expand t in
          O.Chdir (dir, t)))
   | Setenv (var, value, t) ->
@@ -537,7 +531,7 @@ let rec expand (t : Dune_lang.Action.t) : Action.t Action_expander.t =
       then E.consume_file file2
       else
         let+ p = E.dep file2 in
-        as_in_build_dir p ~loc:(String_with_vars.loc file2) ~what:"File"
+        Expander0.as_in_build_dir p ~loc:(String_with_vars.loc file2) ~what:"File"
     in
     O.Diff { optional; file1; file2; mode }
   | No_infer t -> A.no_infer (expand t)
@@ -595,24 +589,24 @@ let expand
     Dep_conf_eval.named ~expander deps_written_by_user
   in
   let expander =
-    match (targets_written_by_user : _ Targets_spec.t) with
-    | Infer -> expander
-    | Static { targets; multiplicity } ->
-      Expander.add_bindings_full
-        expander
-        ~bindings:
-          (Pform.Map.singleton
-             (Var
-                (match multiplicity with
-                 | One -> Target
-                 | Multiple -> Targets))
-             (Expander.Deps.Without
-                (Memo.return
-                   (Value.L.paths
-                      (List.map targets ~f:(fun (target, (_ : Targets_spec.Kind.t)) ->
-                         Path.build target))))))
-  in
-  let expander =
+    let expander =
+      match (targets_written_by_user : _ Targets_spec.t) with
+      | Infer -> expander
+      | Static { targets; multiplicity } ->
+        Expander.add_bindings_full
+          expander
+          ~bindings:
+            (Pform.Map.singleton
+               (Var
+                  (match multiplicity with
+                   | One -> Target
+                   | Multiple -> Targets))
+               (Expander.Deps.Without
+                  (Memo.return
+                     (Value.L.paths
+                        (List.map targets ~f:(fun (target, (_ : Targets_spec.Kind.t)) ->
+                           Path.build target))))))
+    in
     Expander.set_expanding_what expander (User_action targets_written_by_user)
   in
   let+! { Action_builder.With_targets.build; targets } =
