@@ -82,13 +82,6 @@ let gen_rules sctx t ~dir ~scope =
         Some
           (Path.Build.append_source (Super_context.context sctx |> Context.build_dir) p)
       else None)
-  and* prog =
-    Super_context.resolve_program
-      sctx
-      ~dir
-      ~loc:(Some loc)
-      name
-      ~hint:"opam install cinaps"
   in
   let cinaps_dir =
     let stamp =
@@ -122,13 +115,21 @@ let gen_rules sctx t ~dir ~scope =
       sctx
       ~loc:t.loc
       ~dir
-      (Command.run
+      (let prog =
+         Super_context.resolve_program
+           sctx
+           ~dir
+           ~loc:(Some loc)
+           name
+           ~hint:"opam install cinaps"
+       in
+       Command.run_dyn_prog
          ~dir:(Path.build dir)
          prog
          ~sandbox
          [ A "-staged"; Target cinaps_ml; Deps (List.map cinapsed_files ~f:Path.build) ])
   and* expander = Super_context.expander sctx ~dir in
-  let preprocess =
+  let* preprocess =
     Preprocessing.make
       sctx
       ~dir
@@ -188,7 +189,7 @@ let gen_rules sctx t ~dir ~scope =
       cctx
       ~link_args
       ~program:{ name; main_module_name; loc }
-      ~linkages:[ Exe.Linkage.native_or_custom (Super_context.context sctx) ]
+      ~linkages:[ Exe.Linkage.native_or_custom (Compilation_context.ocaml cctx) ]
       ~promote:None
   in
   let action =
