@@ -86,13 +86,18 @@ end = struct
       let open Memo.O in
       Lib.DB.resolve_when_exists lib_db (Loc.none, Lib_name.of_string n)
       >>= (function
-      | Some l -> Resolve.Memo.lift l
-      | None -> resolve_first lib_db l)
+       | Some l -> Resolve.Memo.lift l
+       | None -> resolve_first lib_db l)
   ;;
 end
 
 let coqc ~loc ~dir ~sctx =
-  Super_context.resolve_program sctx "coqc" ~dir ~loc:(Some loc) ~hint:"opam install coq"
+  Super_context.resolve_program_memo
+    sctx
+    "coqc"
+    ~dir
+    ~loc:(Some loc)
+    ~hint:"opam install coq"
 ;;
 
 let select_native_mode ~sctx ~dir (buildable : Coq_stanza.Buildable.t) =
@@ -111,10 +116,10 @@ let select_native_mode ~sctx ~dir (buildable : Coq_stanza.Buildable.t) =
     then Memo.return Coq_mode.VoOnly
     else
       let* coqc = coqc ~sctx ~dir ~loc:buildable.loc in
-      let+ config = Coq_config.make_opt ~coqc in
+      let+ config = Coq_config.make ~coqc in
       (match config with
-       | None -> Coq_mode.VoOnly
-       | Some config ->
+       | Error _ -> Coq_mode.VoOnly
+       | Ok config ->
          (match Coq_config.by_name config "coq_native_compiler_default" with
           | Some (String "yes") | Some (String "ondemand") -> Coq_mode.Native
           | _ -> Coq_mode.VoOnly))
@@ -404,7 +409,7 @@ let setup_coqdep_for_theory_rule
   in
   let stdout_to = dep_theory_file ~dir ~wrapper_name in
   let* coqdep =
-    Super_context.resolve_program
+    Super_context.resolve_program_memo
       sctx
       "coqdep"
       ~dir
@@ -644,31 +649,6 @@ let source_rule ~sctx theories =
      List.concat l)
 ;;
 
-let coqdoc_directory ~mode ~obj_dir ~name =
-  Path.Build.relative
-    obj_dir
-    (Coq_lib_name.to_string name
-     ^
-     match mode with
-     | `Html -> ".html"
-     | `Latex -> ".tex")
-;;
-
-let coqdoc_directory_targets ~dir:obj_dir (theory : Coq_stanza.Theory.t) =
-  let+ (_ : Coq_lib.DB.t) =
-    (* We force the creation of the coq_lib db here so that errors there can
-       appear before any errors to do with directory targets from coqdoc. *)
-    let* scope = Scope.DB.find_by_dir obj_dir in
-    Scope.coq_libs scope
-  in
-  let loc = theory.buildable.loc in
-  let name = snd theory.name in
-  Path.Build.Map.of_list_exn
-    [ coqdoc_directory ~mode:`Html ~obj_dir ~name, loc
-    ; coqdoc_directory ~mode:`Latex ~obj_dir ~name, loc
-    ]
-;;
-
 let setup_coqdoc_rules ~sctx ~dir ~theories_deps (s : Coq_stanza.Theory.t) coq_modules =
   let loc, name = s.buildable.loc, snd s.name in
   let rule =
@@ -684,14 +664,14 @@ let setup_coqdoc_rules ~sctx ~dir ~theories_deps (s : Coq_stanza.Theory.t) coq_m
     fun mode ->
       let* () =
         let* coqdoc =
-          Super_context.resolve_program
+          Super_context.resolve_program_memo
             sctx
             "coqdoc"
             ~dir
             ~loc:(Some loc)
             ~hint:"opam install coq"
         in
-        (let doc_dir = coqdoc_directory ~mode ~obj_dir:dir ~name in
+        (let doc_dir = Coq_doc.coqdoc_directory ~mode ~obj_dir:dir ~name in
          let file_flags =
            let globs =
              let open Action_builder.O in
@@ -752,7 +732,7 @@ let setup_coqdoc_rules ~sctx ~dir ~theories_deps (s : Coq_stanza.Theory.t) coq_m
         | `Html -> Alias.make Alias0.doc ~dir
         | `Latex -> Alias.make (Alias.Name.of_string "doc-latex") ~dir
       in
-      coqdoc_directory ~mode ~obj_dir:dir ~name
+      Coq_doc.coqdoc_directory ~mode ~obj_dir:dir ~name
       |> Path.build
       |> Action_builder.path
       |> Rules.Produce.Alias.add_deps alias ~loc
@@ -1013,7 +993,7 @@ let install_rules ~sctx ~dir s =
 
 let setup_coqpp_rules ~sctx ~dir ({ loc; modules } : Coq_stanza.Coqpp.t) =
   let* coqpp =
-    Super_context.resolve_program
+    Super_context.resolve_program_memo
       sctx
       "coqpp"
       ~dir
