@@ -1,16 +1,16 @@
 (** Loads dune files from the file system.
 
     Also responsible for evaluating dune files written in OCaml syntax. *)
-open! Dune_engine
 
-open! Stdune
+open Import
 
 module Dune_files : sig
   (** A partially evaluated dune file. The context's ocamlc is used to evaluate
       dune files in ocaml syntax *)
   type t
 
-  val eval : t -> context:Context.t -> Dune_file.t list Memo.Build.t
+  val eval : t -> context:Context_name.t -> Dune_file.t list Memo.t
+  val in_dir : Path.Build.t -> Dune_file.t option Memo.t
 end
 
 type conf = private
@@ -20,4 +20,4 @@ type conf = private
   }
 
 (** Load all dune files. This function is memoized. *)
-val load : unit -> conf Memo.Build.t
+val load : unit -> conf Memo.t
