@@ -2,6 +2,10 @@ module Alias = Alias0
 module Findlib = Findlib
 module Main = Main
 module Context = Context
+module Env_node = Env_node
+module Link_flags = Link_flags
+module Ocaml_flags = Ocaml_flags
+module Js_of_ocaml = Js_of_ocaml
 module Super_context = Super_context
 module Compilation_context = Compilation_context
 module Colors = Colors
@@ -56,6 +60,7 @@ module Sub_dirs = Sub_dirs
 module Package = Package
 module Dialect = Dialect
 module Private_context = Private_context
+module Odoc = Odoc
 
 module Install_rules = struct
   let install_file = Install_rules.install_file
