@@ -48,13 +48,13 @@ type t =
   { dir : Path.Build.t
   ; env : Env.t
   ; local_env : string Action_builder.t Env.Var.Map.t
-  ; lib_artifacts : Artifacts.Public_libs.t
-  ; lib_artifacts_host : Artifacts.Public_libs.t
-  ; bin_artifacts_host : Artifacts.Bin.t
+  ; lib_artifacts : Lib.DB.t
+  ; lib_artifacts_host : Lib.DB.t
+  ; artifacts_host : Artifacts.t
   ; bindings : value Pform.Map.t
   ; scope : Scope.t
   ; scope_host : Scope.t
-  ; c_compiler : string
+  ; c_compiler : string Memo.t
   ; context : Context.t
   ; lookup_artifacts : (dir:Path.Build.t -> Ml_sources.Artifacts.t Memo.t) option
   ; foreign_flags :
@@ -63,7 +63,7 @@ type t =
   }
 
 let scope t = t.scope
-let artifacts t = t.bin_artifacts_host
+let artifacts t = t.artifacts_host
 let dir t = t.dir
 let context t = t.context
 let set_foreign_flags t ~f:foreign_flags = { t with foreign_flags }
@@ -74,7 +74,7 @@ let set_local_env_var t ~var ~value =
 
 let set_dir t ~dir = { t with dir }
 let set_scope t ~scope ~scope_host = { t with scope; scope_host }
-let set_bin_artifacts t ~bin_artifacts_host = { t with bin_artifacts_host }
+let set_artifacts t ~artifacts_host = { t with artifacts_host }
 let set_lookup_ml_sources t ~f = { t with lookup_artifacts = Some f }
 let set_expanding_what t x = { t with expanding_what = x }
 
@@ -114,7 +114,7 @@ let dep p =
 let expand_version { scope; _ } ~(source : Dune_lang.Template.Pform.t) s =
   let value_from_version = function
     | None -> [ Value.String "" ]
-    | Some s -> [ String s ]
+    | Some s -> [ String (Package_version.to_string s) ]
   in
   let project = Scope.project scope in
   match
@@ -144,14 +144,14 @@ let expand_version { scope; _ } ~(source : Dune_lang.Template.Pform.t) s =
     let open Memo.O in
     Lib.DB.find (Scope.libs scope) libname
     >>| (function
-    | Some lib -> value_from_version (Lib_info.version (Lib.info lib))
-    | None ->
-      User_error.raise
-        ~loc:source.loc
-        [ Pp.textf
-            "Package %S doesn't exist in the current project and isn't installed either."
-            s
-        ])
+     | Some lib -> value_from_version (Lib_info.version (Lib.info lib))
+     | None ->
+       User_error.raise
+         ~loc:source.loc
+         [ Pp.textf
+             "Package %S doesn't exist in the current project and isn't installed either."
+             s
+         ])
 ;;
 
 let expand_artifact ~source t a s =
@@ -189,19 +189,22 @@ let expand_artifact ~source t a s =
             ~what:"Library"
             (Lib_name.to_string name)
         | Some lib ->
-          let archives = Mode.Dict.get (Lib_info.archives lib) mode in
-          Action_builder.all
-            (List.map archives ~f:(fun fn ->
-               let fn = Path.build fn in
-               let+ () = Action_builder.path fn in
-               Value.Path fn))))
+          Mode.Dict.get (Lib_info.archives lib) mode
+          |> Action_builder.List.map ~f:(fun fn ->
+            let fn = Path.build fn in
+            let+ () = Action_builder.path fn in
+            Value.Path fn)))
 ;;
 
 let cc t =
-  Memo.map (t.foreign_flags ~dir:t.dir) ~f:(fun cc ->
-    Foreign_language.Dict.map cc ~f:(fun cc ->
-      let+ flags = cc in
-      strings (t.c_compiler :: flags)))
+  let make (language : Foreign_language.t) =
+    let+ cc =
+      let* cc = Action_builder.of_memo @@ t.foreign_flags ~dir:t.dir in
+      Foreign_language.Dict.get cc language
+    and+ c_compiler = Action_builder.of_memo t.c_compiler in
+    strings (c_compiler :: cc)
+  in
+  { Foreign_language.Dict.c = make C; cxx = make Cxx }
 ;;
 
 let get_prog = function
@@ -222,6 +225,7 @@ type nonrec expansion_result =
   | Need_full_expander of (t -> value)
 
 let static v = Direct (Without (Memo.return v))
+let static' v = Direct (Without v)
 
 let[@inline never] invalid_use_of_target_variable
   t
@@ -260,7 +264,10 @@ let expand_read_macro ~dir ~source s ~read ~pack =
   let path = relative ~source dir s in
   let read =
     let open Memo.O in
-    let+ x = Build_system.read_file path ~f:read in
+    let+ x =
+      Build_system.read_file path ~f:(fun a -> Async.async (fun () -> read a))
+      >>= Memo.of_reproducible_fiber
+    in
     pack x
   in
   Need_full_expander
@@ -274,22 +281,24 @@ let expand_read_macro ~dir ~source s ~read ~pack =
         With (Action_builder.of_memo read))
 ;;
 
-let file_of_lib { Artifacts.Public_libs.context; public_libs } ~loc ~lib ~file =
+let file_of_lib db context ~loc ~lib ~file =
   let open Resolve.Memo.O in
-  let+ lib = Lib.DB.resolve public_libs (loc, lib) in
-  let dir =
+  let* lib = Lib.DB.resolve db (loc, lib) in
+  let+ dir =
     let info = Lib.info lib in
     match Lib.is_local lib with
-    | false -> Lib_info.src_dir info
+    | false -> Resolve.Memo.return @@ Lib_info.src_dir info
     | true ->
       let name = Lib.name lib in
       let subdir =
         Lib_info.Status.relative_to_package (Lib_info.status info) name
         |> Option.value_exn
       in
-      let pkg_root =
+      let+ pkg_root =
         let package = Lib_name.package_name name in
-        Install.Context.lib_dir ~context:(Context.name context) ~package
+        (* Why do we return the install path? *)
+        let+ context = Resolve.Memo.lift_memo @@ context >>| Context.name in
+        Install.Context.lib_dir ~context ~package
       in
       Path.build (Path.Build.append_local pkg_root subdir)
   in
@@ -329,8 +338,12 @@ let expand_lib_variable t source ~lib ~file ~lib_exec ~lib_private =
                     |> String.quoted)
              ])
     else (
-      let artifacts = if lib_exec then t.lib_artifacts_host else t.lib_artifacts in
-      file_of_lib artifacts ~loc ~lib ~file)
+      let artifacts, context =
+        if lib_exec
+        then t.lib_artifacts_host, Context.host t.context
+        else t.lib_artifacts, Memo.return t.context
+      in
+      file_of_lib artifacts context ~loc ~lib ~file)
   in
   let p =
     let open Memo.O in
@@ -405,8 +418,43 @@ let make loc context =
   [ Value.Path make ]
 ;;
 
+let lib_config_var (var : Pform.Var.t) (lib_config : Lib_config.t) =
+  string
+  @@
+  match var with
+  | Ocaml_stdlib_dir -> Path.to_string lib_config.stdlib_dir
+  | Ext_obj -> lib_config.ext_obj
+  | Ext_lib -> lib_config.ext_lib
+  | Ext_dll -> lib_config.ext_dll
+  | Ccomp_type -> Ocaml_config.Ccomp_type.to_string lib_config.ccomp_type
+  | _ -> Code_error.raise "not a Lib_config.t variable" [ "var", Pform.Var.to_dyn var ]
+;;
+
+let ocaml_config_var (var : Pform.Var.t) (ocaml_config : Ocaml_config.t) =
+  match var with
+  | Cpp -> c_compiler_and_flags ocaml_config @ [ "-E" ] |> strings
+  | Pa_cpp ->
+    c_compiler_and_flags ocaml_config @ [ "-undef"; "-traditional"; "-x"; "c"; "-E" ]
+    |> strings
+  | Arch_sixtyfour -> 64 = Ocaml_config.word_size ocaml_config |> string_of_bool |> string
+  | Ocaml_version -> Ocaml_config.version_string ocaml_config |> string
+  | Ext_asm -> Ocaml_config.ext_asm ocaml_config |> string
+  | Ext_plugin ->
+    (if Ocaml_config.natdynlink_supported ocaml_config then Mode.Native else Byte)
+    |> Mode.plugin_ext
+    |> string
+  | Os_type ->
+    Ocaml_config.os_type ocaml_config |> Ocaml_config.Os_type.to_string |> string
+  | Architecture -> Ocaml_config.architecture ocaml_config |> string
+  | System -> Ocaml_config.system ocaml_config |> string
+  | Model -> Ocaml_config.model ocaml_config |> string
+  | Ext_exe -> Ocaml_config.ext_exe ocaml_config |> string
+  | _ -> Code_error.raise "not a ocaml_config variables" [ "var", Pform.Var.to_dyn var ]
+;;
+
 let expand_pform_var (context : Context.t) ~source (var : Pform.Var.t) =
-  let lib_config = (Context.ocaml context).lib_config in
+  let open Memo.O in
+  let ocaml = Context.ocaml context in
   match var with
   | Pkg _ -> assert false
   | Nothing -> static []
@@ -431,73 +479,55 @@ let expand_pform_var (context : Context.t) ~source (var : Pform.Var.t) =
   | Targets ->
     Need_full_expander
       (fun t -> invalid_use_of_target_variable t ~source ~var_multiplicity:Multiple)
-  | Ocaml -> static (get_prog (Context.ocaml context).ocaml)
-  | Ocamlc -> static (path (Context.ocaml context).ocamlc)
-  | Ocamlopt -> static (get_prog (Context.ocaml context).ocamlopt)
-  | Make -> Direct (Without (make (Dune_lang.Template.Pform.loc source) context))
-  | Cpp ->
-    c_compiler_and_flags (Context.ocaml context).ocaml_config @ [ "-E" ]
-    |> strings
-    |> static
-  | Pa_cpp ->
-    c_compiler_and_flags (Context.ocaml context).ocaml_config
-    @ [ "-undef"; "-traditional"; "-x"; "c"; "-E" ]
-    |> strings
-    |> static
-  | Arch_sixtyfour ->
-    64
-    = Ocaml_config.word_size (Context.ocaml context).ocaml_config
-    |> string_of_bool
-    |> string
-    |> static
-  | Ocaml_bin_dir -> static [ Dir (Context.ocaml context).bin_dir ]
-  | Ocaml_version ->
-    static (string (Ocaml_config.version_string (Context.ocaml context).ocaml_config))
-  | Ocaml_stdlib_dir -> static (string (Path.to_string lib_config.stdlib_dir))
-  | Dev_null -> static (string (Path.to_string Dev_null.path))
-  | Ext_obj -> static (string lib_config.ext_obj)
-  | Ext_asm -> static (string (Ocaml_config.ext_asm (Context.ocaml context).ocaml_config))
-  | Ext_lib -> static (string lib_config.ext_lib)
-  | Ext_dll -> static (string lib_config.ext_dll)
-  | Ext_exe -> static (string (Ocaml_config.ext_exe (Context.ocaml context).ocaml_config))
-  | Ext_plugin ->
-    (if Ocaml_config.natdynlink_supported (Context.ocaml context).ocaml_config
-     then Mode.Native
-     else Byte)
-    |> Mode.plugin_ext
-    |> string
-    |> static
   | Profile -> Context.profile context |> Profile.to_string |> string |> static
   | Workspace_root -> static [ Value.Dir (Path.build (Context.build_dir context)) ]
   | Context_name -> static (string (Context_name.to_string (Context.name context)))
-  | Os_type ->
-    static
-    @@ string
-    @@ Ocaml_config.Os_type.to_string
-         (Ocaml_config.os_type (Context.ocaml context).ocaml_config)
-  | Architecture ->
-    static (string (Ocaml_config.architecture (Context.ocaml context).ocaml_config))
-  | System -> static (string (Ocaml_config.system (Context.ocaml context).ocaml_config))
-  | Model -> static (string (Ocaml_config.model (Context.ocaml context).ocaml_config))
+  | Ocaml ->
+    static'
+    @@ let+ ocaml = ocaml in
+       get_prog ocaml.ocaml
+  | Ocaml_bin_dir ->
+    static'
+    @@ let+ ocaml = ocaml in
+       [ Value.Dir ocaml.bin_dir ]
+  | Ocamlc ->
+    static'
+    @@ let+ ocaml = ocaml in
+       path ocaml.ocamlc
+  | Ocamlopt ->
+    static'
+    @@ let+ ocaml = ocaml in
+       get_prog ocaml.ocamlopt
+  | Make -> Direct (Without (make (Dune_lang.Template.Pform.loc source) context))
+  | Dev_null ->
+    Path.to_string Dev_null.path
+    |> string
+    |> static (* why don't we expand this as a path? *)
+  | Ocaml_stdlib_dir | Ext_obj | Ext_lib | Ext_dll | Ccomp_type ->
+    (let+ ocaml = ocaml in
+     lib_config_var var ocaml.lib_config)
+    |> static'
+  | Ext_exe
+  | Cpp
+  | Pa_cpp
+  | Arch_sixtyfour
+  | Ocaml_version
+  | Ext_asm
+  | Ext_plugin
+  | Os_type
+  | Architecture
+  | System
+  | Model ->
+    (let+ ocaml = ocaml in
+     ocaml_config_var var ocaml.ocaml_config)
+    |> static'
   | Ignoring_promoted_rules ->
     static (string (string_of_bool !Clflags.ignore_promoted_rules))
   | Project_root ->
     Need_full_expander
       (fun t -> Without (Memo.return [ Value.Dir (Path.build (Scope.root t.scope)) ]))
-  | Cc ->
-    Need_full_expander
-      (fun t ->
-        With
-          (let* cc = Action_builder.of_memo (cc t) in
-           cc.c))
-  | Cxx ->
-    Need_full_expander
-      (fun t ->
-        With
-          (let* cc = Action_builder.of_memo (cc t) in
-           cc.cxx))
-  | Ccomp_type ->
-    static @@ string @@ Ocaml_config.Ccomp_type.to_string lib_config.ccomp_type
+  | Cc -> Need_full_expander (fun t -> With (cc t).c)
+  | Cxx -> Need_full_expander (fun t -> With (cc t).cxx)
   | Toolchain ->
     static
     @@ string
@@ -520,20 +550,25 @@ let expand_pform_macro
   | Pkg -> Code_error.raise "pkg forms aren't possible here" []
   | Pkg_self -> Code_error.raise "pkg-self forms aren't possible here" []
   | Ocaml_config ->
-    static
+    static'
     @@
-      (match Ocaml_config.by_name (Context.ocaml context).ocaml_config s with
-      | None ->
-        User_error.raise
-          ~loc:(Dune_lang.Template.Pform.loc source)
-          [ Pp.textf "Unknown ocaml configuration variable %S" s ]
-      | Some v ->
-        (match v with
-         | Bool x -> string (string_of_bool x)
-         | Int x -> string (string_of_int x)
-         | String x -> string x
-         | Words x -> strings x
-         | Prog_and_args x -> strings (x.prog :: x.args)))
+    let open Memo.O in
+    let+ ocaml_config =
+      let+ ocaml = Context.ocaml context in
+      ocaml.ocaml_config
+    in
+    (match Ocaml_config.by_name ocaml_config s with
+     | None ->
+       User_error.raise
+         ~loc:(Dune_lang.Template.Pform.loc source)
+         [ Pp.textf "Unknown ocaml configuration variable %S" s ]
+     | Some v ->
+       (match v with
+        | Bool x -> string (string_of_bool x)
+        | Int x -> string (string_of_int x)
+        | String x -> string x
+        | Words x -> strings x
+        | Prog_and_args x -> strings (x.prog :: x.args)))
   | Env ->
     Need_full_expander
       (fun t ->
@@ -568,9 +603,9 @@ let expand_pform_macro
         With
           (let* prog =
              Action_builder.of_memo
-               (Artifacts.Bin.binary
+               (Artifacts.binary
                   ~loc:(Some (Dune_lang.Template.Pform.loc source))
-                  t.bin_artifacts_host
+                  t.artifacts_host
                   s)
            in
            dep (Action.Prog.ok_exn prog)))
@@ -596,7 +631,7 @@ let expand_pform_macro
       (fun t ->
         Without
           (let open Memo.O in
-           let+ b = Artifacts.Bin.binary_available t.bin_artifacts_host s in
+           let+ b = Artifacts.binary_available t.artifacts_host s in
            b |> string_of_bool |> string))
   | Read -> expand_read_macro ~dir ~source s ~read:Io.read_file ~pack:string
   | Read_lines -> expand_read_macro ~dir ~source s ~read:Io.lines_of_file ~pack:strings
@@ -617,18 +652,31 @@ let expand_pform_macro
       (fun t ->
         Without
           (let open Memo.O in
-           let* coqc = Artifacts.Bin.binary t.bin_artifacts_host ~loc:None "coqc" in
+           let* coqc = Artifacts.binary t.artifacts_host ~loc:None "coqc" in
            let+ t = Coq_config.make ~coqc in
-           match Coq_config.by_name t s with
-           | None ->
+           match t with
+           | Error msg ->
              User_error.raise
                ~loc:(Dune_lang.Template.Pform.loc source)
-               [ Pp.textf "Unknown Coq configuration variable %S" s ]
-           | Some v ->
-             (match v with
-              | Int x -> string (string_of_int x)
-              | String x -> string x
-              | Path x -> Value.L.paths [ x ])))
+               [ Pp.textf "Could not expand %%{coq:%s} as running coqc --config failed." s
+               ; msg
+               ]
+               ~hints:
+                 [ Pp.textf
+                     "coqc --config requires the coq-stdlib package in order to function \
+                      properly."
+                 ]
+           | Ok t ->
+             (match Coq_config.by_name t s with
+              | None ->
+                User_error.raise
+                  ~loc:(Dune_lang.Template.Pform.loc source)
+                  [ Pp.textf "Unknown Coq configuration variable %S" s ]
+              | Some v ->
+                (match v with
+                 | Int x -> string (string_of_int x)
+                 | String x -> string x
+                 | Path x -> Value.L.paths [ x ]))))
 ;;
 
 let expand_pform_gen ~(context : Context.t) ~bindings ~dir ~source (pform : Pform.t)
@@ -725,8 +773,13 @@ let make_root
   ~env
   ~lib_artifacts
   ~lib_artifacts_host
-  ~bin_artifacts_host
+  ~artifacts_host
   =
+  let c_compiler =
+    let open Memo.O in
+    let+ ocaml = Context.ocaml context in
+    Ocaml_config.c_compiler ocaml.ocaml_config
+  in
   { dir = Context.build_dir context
   ; env
   ; local_env = Env.Var.Map.empty
@@ -735,8 +788,8 @@ let make_root
   ; scope_host
   ; lib_artifacts
   ; lib_artifacts_host
-  ; bin_artifacts_host
-  ; c_compiler = Ocaml_config.c_compiler (Context.ocaml context).ocaml_config
+  ; artifacts_host
+  ; c_compiler
   ; context
   ; lookup_artifacts = None
   ; foreign_flags =
