@@ -1,7 +1,6 @@
 (** MDX integration *)
-open! Dune_engine
 
-open Stdune
+open Import
 
 type t
 
@@ -10,10 +9,10 @@ val enabled_if : t -> Blang.t
 type Stanza.t += T of t
 
 (** Generates the rules to handle the given mdx stanza *)
-val gen_rules :
-     t
+val gen_rules
+  :  t
   -> sctx:Super_context.t
   -> dir:Path.Build.t
   -> scope:Scope.t
   -> expander:Expander.t
-  -> unit Memo.Build.t
+  -> unit Memo.t
