@@ -218,13 +218,11 @@ end = struct
         ];
       Memo.return (Error unix_error)
     | Ok dir_contents ->
-      let dir_contents = Fs_cache.Dir_contents.to_list dir_contents in
       let+ files, dirs =
-        Memo.parallel_map dir_contents ~f:(fun (fn, kind) ->
+        Fs_cache.Dir_contents.to_list dir_contents
+        |> Memo.parallel_map ~f:(fun (fn, (kind : File_kind.t)) ->
           let path = Path.Source.relative path fn in
-          if Path.Source.is_in_build_dir path
-          then Memo.return List.Skip
-          else if is_temp_file fn
+          if Path.Source.is_in_build_dir path || is_temp_file fn
           then Memo.return List.Skip
           else
             let+ is_directory, file =
@@ -232,13 +230,13 @@ end = struct
               | S_DIR ->
                 File.of_source_path (In_source_dir path)
                 >>| (function
-                | Ok file -> true, file
-                | Error _ -> true, File.dummy)
+                 | Ok file -> true, file
+                 | Error _ -> true, File.dummy)
               | S_LNK ->
                 Fs_memo.path_stat (In_source_dir path)
                 >>| (function
-                | Ok ({ st_kind = S_DIR; _ } as st) -> true, File.of_stats st
-                | Ok _ | Error _ -> false, File.dummy)
+                 | Ok ({ st_kind = S_DIR; _ } as st) -> true, File.of_stats st
+                 | Ok _ | Error _ -> false, File.dummy)
               | _ -> Memo.return (false, File.dummy)
             in
             if is_directory
