@@ -181,6 +181,7 @@ module Invalidation : sig
       | Event_queue_overflow
       | Upgrade
       | Test
+      | Variable_changed of string
   end
 
   val is_empty : t -> bool
@@ -387,6 +388,26 @@ module With_implicit_output : sig
   val exec : ('i, 'o) t -> 'i -> 'o memo
 end
 
+(** A memoization node that can be explicitly set and invalidated. *)
+module Var : sig
+  type 'a memo := 'a t
+  type 'a t
+
+  (** [create ~name] creates a new variable with the name [name].
+
+      [name]s are currently only used for debugging and there is no validation, i.e., you
+      can create multiple variables with the same name and Memo will not complain. *)
+  val create : ?cutoff:('a -> 'a -> bool) -> 'a -> name:string -> 'a t
+
+  (** [set t value] sets the value of [t] to [value]. Returns the corresponding
+      invalidation that will trigger the subsequent re-evaluation of dependant nodes. *)
+  val set : 'a t -> 'a -> Invalidation.t
+
+  (** [read t] returns the value [t] and introduces a dependency on this node in the
+      current Memo computation. *)
+  val read : 'a t -> 'a memo
+end
+
 (** Memoization of polymorphic functions ['a input -> 'a output t]. The provided
     [id] function must be injective, i.e. there must be a one-to-one
     correspondence between [input]s and their [id]s. *)
@@ -415,43 +436,7 @@ module Debug : sig
   val verbose_diagnostics : bool ref
 end
 
-(** Various performance counters. Reset to zero at the start of every run. *)
-module Perf_counters : sig
-  (** This function must be called to enable performance counters. *)
-  val enable : unit -> unit
-
-  (** Number of nodes restored in the current run. *)
-  val nodes_restored_in_current_run : unit -> int
-
-  (** Number of nodes (re)computed in the current run. *)
-  val nodes_computed_in_current_run : unit -> int
-
-  (** Number of edges traversed in the current run. Some edges may be traversed
-      twice: first when restoring and then when (re)computing a value. *)
-  val edges_traversed_in_current_run : unit -> int
-
-  (** Number of nodes added to the cycle detection DAG in the current run; can't
-      exceed [nodes_restored_in_current_run + nodes_computed_in_current_run]. *)
-  val nodes_for_cycle_detection_in_current_run : unit -> int
-
-  (** Number of edges added to the cycle detection DAG in the current run; can't
-      exceed [edges_traversed_in_current_run]. *)
-  val edges_for_cycle_detection_in_current_run : unit -> int
-
-  (** Number of paths added to the cycle detection DAG in the current run. Each
-      path is a sequence of "forcing" edges followed by a "blocking" edge. *)
-  val paths_for_cycle_detection_in_current_run : unit -> int
-
-  (** A concise summary of performance counters. *)
-  val report_for_current_run : unit -> string
-
-  (** Raise if any internal invariants are violated. *)
-  val assert_invariants : unit -> unit
-
-  (** Reset the counters to zero. You typically don't need to call this function
-      directly (as counters are reset on every run) but it's useful for tests. *)
-  val reset : unit -> unit
-end
+module Metrics = Metrics
 
 module For_tests : sig
   (** After executing a memoized function with a given name and input, it is
